;;;
;;; Copyright (c) 1989, 1993, 1994
;;;	The Regents of the University of California.  All rights reserved.
;;;
;;; This code is derived from software contributed to Berkeley by
;;; Michael Fischbein.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 4. Neither the name of the University nor the names of its contributors
;;;    may be used to endorse or promote products derived from this software
;;;    without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;;; SUCH DAMAGE.
;;;

(.include <sys/cdefs.h> <sys/types.h> <sys/stat.h>
          <ctype.h> <err.h> <fts.h> <limits.h> <stdio.h> <stdlib.h>
          <string.h> <wchar.h> <wctype.h>)
(.include "ls.h" "extern.h")

(define-cise-stmt (make-printer ilseq-handle incomplete-handle
                                nonprintable-handle printable-handle)
  `(let* ([mbs::mbstate_t] [wc::wchar_t] [clen::size_t])
     (memset (& mbs) 0 (sizeof mbs))
     (while (!= (= clen (mbrtowc (& wc) s MB_LEN_MAX (& mbs))) 0)
       (when (== clen (cast size_t -2)) ,incomplete-handle (break))
       (when (== clen (cast size_t -1))
         ,ilseq-handle (memset (& mbs) 0 (sizeof mbs)) (inc! s) (continue))
       (unless (iswprint wc) ,nonprintable-handle (continue))
       ,printable-handle)))

(define-cise-stmt (default-print)
  '(dotimes [i (cast int clen)]
     (putchar (cast (unsigned char) (aref s i)))
     (+= s clen)))

(define-cfn prn_normal (s::(const char*)) ::int
  (let* ([n::int 0])
    (make-printer
     (putchar (cast (unsigned char) (* s)))       ; ilseq (clen == -1)
     (+= n (printf "%s" s))                       ; incomplete (clen == -2)
     (default-print)                              ; nonprintable
     (begin (default-print) (+= n (wcwidth wc)))) ; printable
    (return n)))

(define-cfn prn_printable (s::(const char*)) ::int
  (let* ([n::int 0])
    (make-printer
     (putchar #\?)                                ; ilseq (clen == -1)
     (begin (putchar #\?) (inc! n))               ; incomplete (clen == -2)
     (begin (putchar #\?) (+= s clen) (inc! n))   ; nonprintable
     (begin (default-print) (+= n (wcwidth wc)))) ; printable
    (return n)))
  
;; The fts system makes it difficult to replace fts_name with a different-
;; sized string, so we just calculate the real length here and do the
;; conversion in prn_octal()
;;
;; XXX when using f_octal_escape (-b) rather than f_octal (-B), the
;; length computed by len_octal may be too big. I just can't be buggered
;; to fix this as an efficient fix would involve a lookup table. Same goes
;; for the rather inelegant code in prn_octal.
;;
;;                                              DES 1998/04/23

(define-cfn len_octal (s::(const char*) len::int) ::size_t
  (let* ([r::size_t 0])
    (make-printer
     (begin (+= r 4) (when (== (dec! len) 0) (break))) ; ilseq (clen == -1)
     (+= r (* 4 len))                                  ; incomplete (clen == -2)
     (begin (+= r (* 4 clen)) (+= s clen))             ; nonprintable
     (begin (inc! r) (+= s clen)))                     ; printable
    (return r)))

(define-cise-stmt (octal-print prtlen)
  `(dotimes [i ,prtlen]
     (let* ([ch::(unsigned char) (cast (unsigned char) (aref s i))])
       (putchar #\\)
       (putchar (+ #\0 (>> ch 6)))
       (putchar (+ #\0 (logand (>> ch 3) 7)))
       (putchar (+ #\0 (logand ch 7)))
       (+= len 4))))

(define-cise-stmt (esc-print)
  `(begin
     (if (and f_octal_escape (>= wc 0) (<= wc (cast wchar_t UCHAR_MAX)))
       (let* ([p::(const char*) (strchr esc (cast char wc))])
         (cond [(!= p NULL) (putchar #\\) (putchar (aref p 1)) (+= len 2)]
               [else (octal-print clen)]))
       (octal-print clen))
     (+= s clen)))

(define-cfn prn_octal (s::(const char*)) ::int
  (let* ([len::int 0]
         [esc::(.array (static const char) (())) "\\\\\"\"\aa\bb\ff\nn\rr\tt\vv"])
    (make-printer
     (octal-print 1)                    ; ilseq (clen == -1)
     (octal-print (strlen s))           ; incomplete (clen == -2)
     (esc-print)                        ; nonprint
     (if (and (!= wc (cast wchar_t #\")) (!= wc (cast wchar_t #\\))) ;print
       (begin (default-print) (+= len (wcwidth wc)))
       (esc-print)))
    (return len)))

(define-cfn usage (||::void) ::void
  (fprintf stderr
           "usage: ls [-ABCFHILPRSTUWZabcdfghiklmnopqrstuwx1] [file ...]\n")
  (exit 1))
